import pytest

from synthegrator.lang_specs.lang_spec import (
    PythonFunctionParser,
)
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.lang_specs.lang_util import truncate_when_decreases_indent


@pytest.fixture
def example_dypy_method_nodocstring():
    return """def add_task(
        self,
        task: Task,
        queue: None | BaseTaskQueue = None,
        raise_error: bool = False,
    ) -> bool:
        if queue is None:
            queue = self.task_queue
        if task.priority is None or not task.priority_set_explicitly:
            task.priority = self.generate_task_priority()
            task.priority_set_explicitly = False
        else:
            task.priority_set_explicitly = True

        if not task.request.url or not task.request.url.startswith(
            ("http://", "https://", "ftp://", "file://", "feed://")
        ):
            self.collect_runtime_event("task-with-invalid-url", task.request.url)
            msg = "Invalid task URL: %s" % task.request.url
            if raise_error:
                raise SpiderError(msg)
            logger.error(
                "%s\nTraceback:\n%s",
                msg,
                "".join(format_stack()),
            )
            return False
        # TODO: keep original task priority if it was set explicitly
        # WTF the previous comment means?
        queue.put(task, priority=task.priority, schedule_time=task.schedule_time)
        return True"""


def test_get_prefixcomments_only_at_start():
    langspec_py = PythonLangSpec()
    comments = list(
        langspec_py.get_prefix_comments(
            "def foo():\n"
            "    this_isnot()\n"
            "    '''\n"
            "    This is not a docstring\n"
            "    '''\n"
            "    pass\n"
            "def bar():\n"
            "    this_isnot()\n"
            "    '''\n"
            "    This is not another docstring\n"
            "    '''\n"
            "    pass\n"
            "def boo():\n"
            "    '''\n"
            "    This is a docstring\n"
            "    '''\n"
            "    pass\n",
        ),
    )
    assert len(comments) == 1
    assert comments[0].doc_string_str == "'''\n    This is a docstring\n    '''"


def test_treesitter_getdocstring_only_start1(example_dypy_method_nodocstring):
    parser = PythonFunctionParser(example_dypy_method_nodocstring)
    assert parser.docstring_node is None


def test_treesitter_getdocstring_only_start2(example_dypy_method_nodocstring):
    parser = PythonFunctionParser(
        "def foo():\n"
        "    this_isnot()\n"
        "    '''\n"
        "    This is not a docstring\n"
        "    '''\n"
        "    pass\n",
    )
    assert parser.docstring_node is None


def test_treesitter_getdocstring_only_start3(example_dypy_method_nodocstring):
    parser = PythonFunctionParser(
        "def foo():\n    '''\n    This is a docstring\n    '''\n    pass\n",
    )
    assert (
        parser.docstring_node.text.decode() == "'''\n    This is a docstring\n    '''"
    )


def test_treesitter_getdocstring_only_start1():
    fixture1 = (
        "def foo():\n"
        "    this_isnot()\n"
        "    '''\n"
        "    This is a docstring\n"
        "    '''\n"
        "    pass\n"
    )
    parser = PythonFunctionParser(fixture1)
    assert parser.docstring_node is None


def test_treesitter_getdocstring_only_start2():
    fixture1 = "def foo():\n    '''\n    This is a docstring\n    '''\n    pass\n"
    parser = PythonFunctionParser(fixture1)
    assert parser.docstring_node is not None
    assert (
        parser.docstring_node.text.decode() == "'''\n    This is a docstring\n    '''"
    )


def test_get_docstrings():
    langspec_py = PythonLangSpec()
    comments = list(
        langspec_py.get_prefix_comments(
            "def foo():\n"
            "    '''\n"
            "    This is a docstring\n"
            "    '''\n"
            "    pass\n"
            "def bar():\n"
            "    '''\n"
            "    This is other docstring\n"
            "    '''\n"
            "    pass\n",
        ),
    )
    assert len(comments) == 2
    assert comments[0].doc_string_node.start_point == (1, 4)
    assert comments[0].doc_string_node.end_point == (3, 4 + 3)
    assert (
        comments[0].doc_string_node.text.decode("utf-8")
        == "'''\n    This is a docstring\n    '''"
    )
    assert comments[0].function_node.start_point == (0, 0)
    assert comments[0].function_node.end_point == (4, 4 + 4)


def test_truncate_when_decrease_indent():
    result = truncate_when_decreases_indent("    return 2\nassert False")
    assert result == "    return 2"


def test_truncate_when_decrease_indent2():
    result = truncate_when_decreases_indent(  # fmt: off
        "    x = 2\n"
        "    y = x * 2\n"
        "    for foo in bar:\n"
        "        y += 1\n"
        "    return y\n"
        "assert False",
    )
    assert (
        result
        == "    x = 2\n    y = x * 2\n    for foo in bar:\n        y += 1\n    return y"
    )  # fmt: on
